# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from mo_dots import Null

from mo_json import JX_IS_NULL
from mo_streams.type_utils import Typer

from mo_streams.object_stream import ObjectStream
from mo_streams._utils import Stream


class EmptyStream(Stream):
    def to_dict(self):
        return {}

    def to_list(self):
        return []

    def to_bytes(self):
        return b""

    def to_str(self):
        return ""

    def first(self):
        return None

    def append(self, value):
        def read():
            yield value, {}

        return ObjectStream(read(), Typer(example=value), JX_IS_NULL)

    def to_data(self):
        return Null

    def sum(self):
        return None

    def last(self):
        return None


def return_self(self, *args, **kwargs):
    return self


for prop in vars(ObjectStream):
    if hasattr(EmptyStream, prop):
        continue
    setattr(EmptyStream, prop, return_self)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from mo_dots import Data
from mo_dots.lists import is_many, is_finite
from mo_files import File
from mo_future import first
from mo_imports import export

from mo_json import JxType, JX_TEXT
from mo_streams._utils import Stream, Reader
from mo_streams.byte_stream import ByteStream
from mo_streams.empty_stream import EmptyStream
from mo_streams.files import content, File_usingStream
from mo_streams.function_factory import it
from mo_streams.object_stream import ObjectStream, ERROR, WARNING, NONE
from mo_streams.string_stream import StringStream
from mo_streams.type_utils import Typer, CallableTyper, StreamTyper, LazyTyper


def stream(value):
    if isinstance(value, (dict, Data)):
        kv = first(value.items())
        if not kv:
            return EmptyStream()
        _, example = kv
        return ObjectStream(((v, {"key": k}) for k, v in value.items()), Typer(example=example), JxType(key=JX_TEXT),)
    elif isinstance(value, bytes):
        return ByteStream(Reader(iter([value])))
    elif isinstance(value, str):
        return StringStream(iter([value]))
    elif value == None:
        return EmptyStream
    elif isinstance(value, Stream):
        return value
    elif isinstance(value, type(range(1))):
        return ObjectStream(((v, {}) for v in value), Typer(example=value.stop), JxType())
    elif is_finite(value):
        example = first(value)

        def read_from_list():
            for v in value:
                yield v, {}

        return ObjectStream(read_from_list(), Typer(example=example), JxType())
    elif is_many(value):
        example = first(value)

        def read():
            yield example, {}
            for v in value:
                yield v, {}

        return ObjectStream(read(), Typer(example=example), JxType())
    else:
        return ObjectStream(iter([(value, {})]), Typer(example=value), JxType())


def from_s3(bucket, key):
    import boto3
    return S3Object(boto3.resource('s3').Object(bucket, key))


class S3Object:
    def __init__(self, obj):
        self.obj = obj

    def content(self) -> ByteStream:
        return ByteStream(self.obj.get()['Body'])


STR_CALL = CallableTyper(return_type=str)
BOOL_CALL = CallableTyper(return_type=bool)
INT_CALL = CallableTyper(return_type=int)
BYTES_CALL = CallableTyper(return_type=bytes)
STR = Typer(python_type=str)


ANNOTATIONS = {
    # THE str METHODS DO NOT APPEAR TO BE IN THE str CLASS ANNOTATIONS
    (str, "encode"): BYTES_CALL,
    (str, "startswith"): BOOL_CALL,
    (str, "__repr__"): STR_CALL,
    (str, "__str__"): STR_CALL,
    (str, "__lt__"): BOOL_CALL,
    (str, "__le__"): BOOL_CALL,
    (str, "__eq__"): BOOL_CALL,
    (str, "__ne__"): BOOL_CALL,
    (str, "__gt__"): BOOL_CALL,
    (str, "__ge__"): BOOL_CALL,
    (str, "__len__"): INT_CALL,
    (str, "__add__"): STR_CALL,
    (str, "__contains__"): BOOL_CALL,
    (str, "encode"): BYTES_CALL,
    (str, "replace"): STR_CALL,
    (str, "split"): CallableTyper(return_type=StreamTyper(member_type=STR, _schema=JX_TEXT)),
    (str, "rsplit"): CallableTyper(return_type=StreamTyper(member_type=STR, _schema=JX_TEXT)),
    (str, "join"): STR_CALL,
    (str, "capitalize"): STR_CALL,
    (str, "casefold"): STR_CALL,
    (str, "title"): STR_CALL,
    (str, "center"): STR_CALL,
    (str, "count"): INT_CALL,
    (str, "expandtabs"): STR_CALL,
    (str, "find"): INT_CALL,
    (str, "partition"): BOOL_CALL,
    (str, "index"): INT_CALL,
    (str, "ljust"): STR_CALL,
    (str, "lower"): STR_CALL,
    (str, "lstrip"): STR_CALL,
    (str, "rfind"): INT_CALL,
    (str, "rindex"): INT_CALL,
    (str, "rjust"): BOOL_CALL,
    (str, "rstrip"): BOOL_CALL,
    (str, "rpartition"): BOOL_CALL,
    (str, "splitlines"): BOOL_CALL,
    (str, "strip"): STR_CALL,
    (str, "swapcase"): STR_CALL,
    (str, "translate"): STR_CALL,
    (str, "upper"): STR_CALL,
    (str, "startswith"): BOOL_CALL,
    (str, "endswith"): BOOL_CALL,
    (str, "removeprefix"): STR_CALL,
    (str, "removesuffix"): STR_CALL,
    (str, "isascii"): BOOL_CALL,
    (str, "islower"): BOOL_CALL,
    (str, "isupper"): BOOL_CALL,
    (str, "istitle"): BOOL_CALL,
    (str, "isspace"): BOOL_CALL,
    (str, "isdecimal"): BOOL_CALL,
    (str, "isdigit"): BOOL_CALL,
    (str, "isnumeric"): BOOL_CALL,
    (str, "isalpha"): BOOL_CALL,
    (str, "isalnum"): BOOL_CALL,
    (str, "isidentifier"): BOOL_CALL,
    (str, "isprintable"): BOOL_CALL,
    (File_usingStream, "content"): CallableTyper(return_type=ByteStream),
    (File, "content"): CallableTyper(return_type=ByteStream),
    (ByteStream, "utf8"): CallableTyper(return_type=StringStream),
    (StringStream, "lines"): CallableTyper(return_type=StreamTyper(member_type=STR, _schema=JxType())),
    (ByteStream, "lines"): CallableTyper(return_type=StreamTyper(member_type=STR, _schema=JxType())),
    (ObjectStream, "map"): CallableTyper(return_type=StreamTyper(member_type=LazyTyper(), _schema=JxType())),
}

export("mo_streams.object_stream", stream)
export("mo_streams.type_utils", ANNOTATIONS)
export("mo_streams.function_factory", ANNOTATIONS)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import os
from unittest import TestCase, skipIf, skip

import boto3
from mo_dots import exists, Data
from mo_files import File, TempFile
from mo_logs import logger
from mo_math import randoms
from mo_testing.fuzzytestcase import add_error_reporting
from mo_threads import start_main_thread, stop_main_thread
from mo_times import Date, YEAR
from moto import mock_s3

from mo_json import json2value
from mo_streams import stream, it, ANNOTATIONS, Typer, EmptyStream, from_s3
from mo_streams._utils import Writer
from mo_streams.files import File_usingStream
from mo_streams.string_stream import line_terminator

IS_TRAVIS = bool(os.environ.get("TRAVIS"))


@add_error_reporting
class TestStream(TestCase):

    @classmethod
    def setUpClass(cls):
        stop_main_thread()
        start_main_thread()

    @classmethod
    def tearDownClass(cls):
        stop_main_thread()


    def test_encode(self):
        "".encode("utf8")
        result = stream("this is a test").encode("utf8").to_bytes()
        self.assertEqual(result, b"this is a test")

    def test_zst(self):
        file = File("tests/resources/test.zst")
        stream("this is a test").utf8().to_zst().write(file)
        result = file.content().utf8().to_str()
        self.assertEqual(result, "this is a test")

    @skipIf(IS_TRAVIS, "file not repo")
    def test_tar(self):
        file = File("tests/resources/so_queries.tar.zst")
        content = file.content().content().exists().utf8().to_str().to_list()
        self.assertEqual(len(content), 6191)

    def test_work_on_files(self):
        def analysis(lines):
            ## expect list of Data
            if not isinstance(lines, list):
                logger.error("expecting list")
            if not all(isinstance(d, Data) for d in lines):
                logger.error("expecting Data")
            return lines

        result = (
            File("tests/resources/data_in_zip.zip")
            .content()
            .map(it.content().lines().filter(exists).map(json2value).to_list())
            .map(analysis)
            .to_list()
        )
        self.assertEqual(result, [[{"a": 1}, {"a": 2}, {"a": 3}], [{"c": 11}, {"c": 22}, {"c": 33}]])

    def test_work_on_files_simple(self):
        result = File("tests/resources/data_in_zip.zip").content().map(it.content().lines().to_list()).to_list()
        self.assertEqual(
            result,
            [
                ['{"a":  1}', "", '{"a":  2}', "", '{"a":  3}', "", ""],
                ['{"c":  11}', '{"c":  22}', "", '{"c":  33}', "", ""],
            ],
        )

    def test_from_zip(self):
        file = File("tests/resources/example.zip")
        content = file.content().rel_path.to_list()
        self.assertEqual(content, ["LICENSE", "README.md"])

    def test_dict_zip(self):
        values = ["a", "b", "c"]
        result = stream(values).enumerate().to_dict()
        self.assertEqual(result, {0: "a", 1: "b", 2: "c"})

    def test_to_zip(self):
        with TempFile(f"delete_{randoms.base64(5)}.zip") as file:
            stream(File("tests").leaves).to_zip().write(file)

    def test_lambda(self):
        def func(x):
            return x + 2

        result = stream([1, 2, 3]).map(func).to_list()
        self.assertEqual(result, [3, 4, 5])

    def test_sort(self):
        result = stream([2, 3, 1]).sort().to_list()
        self.assertEqual(result, [1, 2, 3])

    def test_reverse(self):
        result = stream([2, 3, 1]).reverse().to_list()
        self.assertEqual(result, [1, 3, 2])

    def test_distinct(self):
        result = stream([2, 2, 2, 1, 4, 3, 1]).distinct().to_list()
        self.assertEqual(result, [2, 1, 4, 3])

    def test_limit_under(self):
        result = stream(range(200)).limit(10).to_list()
        self.assertEqual(result, list(range(10)))

    def test_limit_over(self):
        result = stream(range(5)).limit(10).to_list()
        self.assertEqual(result, list(range(5)))

    def test_limit_match(self):
        result = stream(range(10)).limit(10).to_list()
        self.assertEqual(result, list(range(10)))

    @skipIf(IS_TRAVIS, "pandas too hard for travis")
    def test_to_zip2(self):
        from pandas import DataFrame

        result = (
            stream({"data": [{"a": 1, "b": 2}]})
            .map(DataFrame)
            .attach(writer=it(Writer)())
            .map(it.to_csv(it.writer))
            .attach(name="test_" + it.key)
            .map(it(File_usingStream)(it.name, it.writer.content))
            .to_zip()
            .write("tests/resources/test_to_zip.zip")
        )

    @skipIf(IS_TRAVIS, "pandas too hard for travis")
    @mock_s3
    def test_zip_to_s3(self):
        from pandas import DataFrame

        s3 = boto3.resource("s3")
        bucket_name = "bucket-" + randoms.hex(10)
        bucket = s3.create_bucket(Bucket=bucket_name)

        result = (
            stream({"data": [{"a": 1, "b": 2}]})
            .map(DataFrame)
            .attach(writer=it(Writer)())
            .map(it.to_csv(it.writer, index=False, **{line_terminator: "\n"}))
            .attach(name="test_" + it.key)
            .map(it(File_usingStream)(it.name, it.writer.content))
            .to_zip()
            .to_s3(name="test", bucket=bucket_name)
        )

        # bucket = s3.Bucket(bucket_name)
        for obj in bucket.objects.all():
            key, body = obj.key, obj.get()["Body"].read()
            if key == "test":
                content = stream(body).from_zip().content().utf8().to_str().to_list()
                self.assertEqual(content, ["a,b\n1,2\n"])

    @skip("not sure why this is an error")
    def test_missing_lambda_parameter(self):
        with self.assertRaises(Exception):
            result = (
                stream({"data": [{"a": 1, "b": 2,}]})
                .attach(writer=lambda: Writer())
                .map(it.writer.write(str(it)))
                .to_list()
            )
            print(result)

    def test_inequality(self):
        self.assertEqual(stream([1, 2, 3]).filter(it > 2).to_list(), [3])
        self.assertEqual(stream([1, 2, 3]).filter(it >= 2).to_list(), [2, 3])
        self.assertEqual(stream([1, 2, 3]).filter(it < 2).to_list(), [1])
        self.assertEqual(stream([1, 2, 3]).filter(it <= 2).to_list(), [1, 2])

    def test_map_int(self):
        result = stream(["1", "2", "3"]).map(int).to_list()
        self.assertEqual(result, [1, 2, 3])

    def test_map_function(self):
        result = stream(["1", "2", "3"]).map(length).to_list()
        self.assertEqual(result, [1, 1, 1])

    def test_map_lambda(self):
        result = stream(["1", "2", "3"]).map(lambda v: int(v) + 1).to_list()
        self.assertEqual(result, [2, 3, 4])

    def test_filter(self):
        result = stream([1, 2, 3]).filter(lambda v: v % 2).to_list()
        self.assertEqual(result, [1, 3])

    def test_group1(self):
        result = (
            stream([1, 2, 3])
            .group(lambda v: v % 2)
            .map(lambda v, att: {"group": att["group"], "value": v.to_list()})
            .to_list()
        )
        self.assertEqual(result, [{"group": 0, "value": [2]}, {"group": 1, "value": [1, 3]}])

    def test_group2(self):
        result = (
            stream([1, 2, 3])
            .group(lambda v: v % 2)
            .map(it.sum())
            .map(lambda v, att: {"group": att["group"], "value": v})
            .to_list()
        )
        self.assertEqual(result, [{"group": 0, "value": 2}, {"group": 1, "value": 4}])

    def test_first(self):
        result = stream([1, 2, 3]).first()
        self.assertEqual(result, 1)

    def test_map_it(self):
        class SomeClass:
            num = 0

            def __init__(self):
                self.value = SomeClass.num
                SomeClass.num += 1

        ANNOTATIONS[(SomeClass, "value")] = Typer(python_type=int)
        result = stream([SomeClass(), SomeClass(), SomeClass()]).map(it.value).last()
        self.assertEqual(result, 2)

    @skip("not implemented yet")
    def test_deep_iteration(self):
        class Something:
            def __init__(self, value):
                self.value = value

        value = Something({"props": [{"a": 1}, {"a": 2}, {"a": 3}]})
        result = stream(value).value.props.a.to_list()
        self.assertEqual(result, [1, 2, 3])

    @skip("not implemented yet")
    def test_reverse_dict(self):
        data = {1: "a", 2: "b", 3: "c", 4: "a"}
        result = stream(data).group(it).map(it.key.to_list()).to_dict(key="group")
        self.assertEqual(result, {"a": [1, 4], "b": [2], "c": ["3"]})

    @skip("not implemented yet")
    def test_pivot(self):
        populations = [
            {"date": "2019-07-01", "cohort": "00", "population": 1000},
            {"date": "2019-07-01", "cohort": "45", "population": 2000},
            {"date": "2020-07-01", "cohort": "00", "population": 3000},
            {"date": "2020-07-01", "cohort": "45", "population": 4000},
            {"date": "2021-07-01", "cohort": "00", "population": 5000},
            {"date": "2021-07-01", "cohort": "45", "population": 6000},
        ]

        last_year = "2020-07-01"
        this_year = "2021-07-01"
        next_year = (Date(this_year) + YEAR).format("%Y-%m-%d")
        years = [this_year, last_year, next_year]
        result = (
            stream(populations)
            # select records we need
            .filter(it.date in years)
            # transpose (pivot date to rows)
            .pivot(values="population", index="cohort", columns="date")
            # calc
            .attach(2 * it[this_year] - it[last_year], next_year)
            # melt dates back to rows
            .melt(id_vars=["cohort"], value_vars=years, var_name="date", value_name="population",)
            # add to original table
            .append_to(populations)
        )
        expected = [
            {"date": "2019-07-01", "cohort": "00", "population": 1000},
            {"date": "2019-07-01", "cohort": "45", "population": 2000},
            {"date": "2020-07-01", "cohort": "00", "population": 3000},
            {"date": "2020-07-01", "cohort": "45", "population": 4000},
            {"date": "2021-07-01", "cohort": "00", "population": 5000},
            {"date": "2021-07-01", "cohort": "45", "population": 6000},
            {"date": "2022-07-01", "cohort": "00", "population": 7000},
            {"date": "2022-07-01", "cohort": "45", "population": 8000},
        ]
        self.assertEqual(result, expected)

    def test_empty_attach(self):
        stream = EmptyStream()
        self.assertIsNone(stream.attach(a=22).sum())

    def test_div(self):
        self.assertEqual(stream([1, 2, None, 3]).map(it / 100).to_list(), [0.01, 0.02, None, 0.03])

    def test_radd(self):
        self.assertEqual(stream([1, 2, None, 3]).map(3 + it).to_list(), [4, 5, None, 6])

    def test_rsub(self):
        self.assertEqual(stream([1, 2, None, 3]).map(3 - it).to_list(), [2, 1, None, 0])

    def test_str_startswith(self):
        self.assertEqual(stream(["abc", "def", "ghi"]).filter(it.startswith("a")).to_list(), ["abc"])

    def test_filter_and_concat_naive(self):
        self.assertEqual(stream(["A", "", None, "d"]).filter(exists).map(str.lower).join(","), "a,d")

    def test_filter_and_concat_simpler(self):
        self.assertEqual(stream(["A", "", None, "d"]).exists().lower().join(","), "a,d")

    @mock_s3
    def test_s3_csv_to_dict(self):
        s3 = boto3.resource("s3")
        bucket_name = "bucket-" + randoms.hex(10)
        bucket = s3.create_bucket(Bucket=bucket_name)

        content = "id,name\n1,Alice\n2,Bob"
        bucket.put_object(Key='test.csv', Body=content)

        result = (
            from_s3(key="test.csv", bucket=bucket_name)
            .content()
            .utf8()
            .csv()
            .to_list()
        )

        self.assertEqual(result, [{"id": "1", "name": "Alice"}, {"id": "2", "name": "Bob"}])

    def test_from_csv_w_map(self):
        result = (
            stream("id,name\n1,Alice\n2,Bob")
            .csv()
            .map(it['id'] + it['name'])
            .to_list()
        )

        self.assertEqual(result, ["1Alice", "2Bob"])

    def test_count_values_and_sort(self):
        values = [
            "apple",
            "pear",
            "apple",
            "orange",
            "pear",
            "apple"
        ]

        counts = (
            stream(values)
            .group(key=it)
            .map(it.count())
            .sort(reverse=True)
            .map(lambda v, att: {att["key"]: v})
            .to_list()
        )

        self.assertEqual(counts, [{"apple": 3}, {"pear": 2}, {"orange": 1}])

    def test_count_values_and_sort2(self):
        values = [
            "apple",
            "pear",
            "apple",
            "orange",
            "pear",
            "apple"
        ]

        counts = (
            stream(values)
            .group(key=it)
            .map(it.count())
            .to_dict()
        )

        self.assertEqual(counts, {"apple": 3, "pear": 2, "orange": 1})


def length(value):
    return len(value)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import inspect

from mo_imports import expect, export
from mo_logs import logger

from mo_json import JxType, JX_TEXT, array_of, JX_IS_NULL
from mo_streams._utils import arg_spec

parse, ANNOTATIONS, ObjectStream = expect("parse", "ANNOTATIONS", "ObjectStream")


class Typer:
    """
    Smooth out the lumps of Python type manipulation
    """

    def __init__(self, *, example=None, python_type=None, function=None):
        if function:
            # find function return type
            inspect.signature(function)
        elif example:
            self.python_type = type(example)
        elif python_type is LazyTyper:
            self.__class__ = LazyTyper
        else:
            self.python_type = python_type

    def __getattr__(self, item):
        try:
            attribute_type = self.python_type.__annotations__[item]
            return Typer(python_type=attribute_type)
        except:
            pass

        desc = arg_spec(self.python_type, item)
        if desc:
            return_type = desc.annotations.get("return")
            if return_type:
                return parse(return_type)
        return_type = ANNOTATIONS.get((self.python_type, item))
        if return_type:
            return return_type

        return UnknownTyper(lambda t: logger.error(
            """expecting {{type}} to have attribute {{item|quote}} declared with a type annotation""",
            type=self.python_type.__name__,
            item=item,
        ))

    def __add__(self, other):
        if self.python_type is str or other.typer is str:
            return Typer(python_type=str)
        if self.python_type in [int, float]:
            return Typer(python_type=float)
        logger.error("not handled")

    def __call__(self, *args, **kwargs):
        logger.error("programmer error")

    def __str__(self):
        return f"Typer(class={self.python_type.__name__})"


class JxTyper(Typer):
    """
    represent Data schema
    """

    def __init__(self, type_):
        self.type_: JxType = type_

    def __getattr__(self, item):
        attribute_type = self.type_[item]
        return Typer(python_type=attribute_type)

        # logger.error(
        #     """expecting {{type}} to have attribute {{item|quote}}""",
        #     type=self.type_,
        #     item=item,
        # )

    __getitem__ = __getattr__

    def __add__(self, other):
        if self.type_ != other.typer:
            logger.error("Can not add two different types")
        if self.type_ == JX_TEXT:
            # ADDING STRINGS RESULTS IN AN ARRAY OF STRINGS
            return array_of(JX_TEXT)
        return self

    def __call__(self, *args, **kwargs):
        return JX_IS_NULL

    def __str__(self):
        return f"JxTyper({self.type_.__repr__()})"

    __repr__ = __str__


class StreamTyper(Typer):
    """
    AN ObjectStream HAS A TYPE TOO
    """

    def __init__(self, member_type, _schema):
        Typer.__init__(self, python_type=ObjectStream)
        if not isinstance(member_type, Typer):
            logger.error("expecting typer")
        self.member_type = member_type
        self._schema = _schema

    def __call__(self, *args, **kwargs):
        logger.error("can not call an ObjectStream")

    @property
    def map(self):
        return MapperTyper(self.member_type, self._schema)

    @property
    def filter(self):
        return CallableTyper(self)

    @property
    def to_list(self):
        return CallableTyper(Typer(python_type=list))

    @property
    def sum(self):
        return CallableTyper(Typer(python_type=float))

    @property
    def count(self):
        return CallableTyper(Typer(python_type=int))

    def __getattr__(self, item):
        spec = inspect.getmembers(ObjectStream)
        for k, m in spec:
            if k == item:
                logger.error("add method (above) to handle type inference for ObjectStream.{item}", item=item)

        output = getattr(self.member_type, item)
        if isinstance(output, UnknownTyper):
            if item in self._schema:
                output = self._schema[item]
                if output:
                    return JxTyper(output)
        return output

    def __str__(self):
        return f"StreamTyper({self.member_type})"


class MapperTyper(Typer):
    """
    REPRESENT THE RETURN TYPE OF THE Stream.map()
    """

    def __init__(self, domain_type, _schema):
        Typer.__init__(self, python_type=ObjectStream)
        self.domain_type = domain_type
        self._schema = _schema

    def __call__(self, return_type):
        return StreamTyper(member_type=return_type, _schema=self._schema)

    def __str__(self):
        return f"MapperTyper({self.member_type}, {self._schema})"


class CallableTyper(Typer):
    """
    ASSUME THIS WILL BE CALLED, AND THIS IS THE TYPE RETURNED
    """

    def __init__(self, return_type):
        if isinstance(return_type, Typer):
            self.type_ = return_type
        else:
            self.type_ = Typer(python_type=return_type)

    def __call__(self, *args, **kwargs):
        return self.type_

    def __getattr__(self, item):
        spec = inspect.getmembers(self.type_)
        for k, m in spec:
            if k == item:
                inspect.ismethod(m)

    def __str__(self):
        return f"CallableTyper(return_type={self.type_.__repr__})"

    __repr__ = __str__


class UnknownTyper(Typer):
    """
    MANY TIMES WE DO NOT KNOW THE TYPE, BUT MAYBE WE NEVER NEED IT
    """

    def __init__(self, error):
        Typer.__init__(self)
        self._error: Exception = error

    def __bool__(self):
        return False

    def __getattr__(self, item):
        def build(type_):
            return getattr(type_, item)

        return UnknownTyper(build)

    def __call__(self, *args, **kwargs):
        def build(type_):
            return type_()

        return UnknownTyper(build)

    def __str__(self):
        return "UnknownTyper()"


class LazyTyper(Typer):
    """
    PLACEHOLDER FOR STREAM ELEMENT TYPE, UNKNOWN DURING LAMBDA DEFINITION
    """

    def __init__(self, resolver=None):
        Typer.__init__(self)
        self._resolver = resolver or (lambda t: t)

    def __getitem__(self, item):
        if not isinstance(item, LazyTyper):
            other = LazyTyper(lambda t: other)

        def build(type_):
            return type_[item]

        return LazyTyper(build)

    def __getattr__(self, item):
        def build(type_):
            return getattr(type_, item)

        return LazyTyper(build)

    def __call__(self, *args, **kwargs):
        def build(type_):
            return type_

        return LazyTyper(build)

    def __add__(self, other):
        if not isinstance(other, LazyTyper):
            other = LazyTyper(lambda t: other)

        def build(type_):
            return self._resolver(type_) + other._resolver(type_)

        return LazyTyper(build)

    def __str__(self):
        return "LazyTyper()"


export("mo_streams.type_parser", Typer)
export("mo_streams.byte_stream", Typer)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import inspect
from collections import namedtuple
from types import FunctionType

from mo_dots import is_missing
from mo_future import first
from mo_imports import expect
from mo_logs import logger, Except, strings
from mo_logs.exceptions import ERROR, get_stacktrace

from mo_streams.type_utils import Typer, LazyTyper, CallableTyper, UnknownTyper

DEBUG = False

ANNOTATIONS = expect('ANNOTATIONS')


_get = object.__getattribute__
_set = object.__setattr__

BuiltFunction = namedtuple("BuiltFunction", ["function", "return_type", "schema"])
NO_ARGS = BuiltFunction(tuple(), tuple(), tuple())


class FunctionFactory:
    """
    See mo-streams/docs/function_factory.md
    """

    def __init__(self, builder, typer, desc):
        if not isinstance(typer, Typer):
            logger.error("expecting type, not {{type}}", type=typer)
        _set(self, "build", builder)
        _set(self, "typer", typer)
        _set(self, "_desc", desc)

    def __getitem__(self, item):
        source = f"{self}[{item}]"

        def builder(domain_type, domain_schema) -> BuiltFunction:
            f, t, s = _get(self, "build")(domain_type, domain_schema)
            if item in s:

                def get_schema_item(v, a):
                    try:
                        return a[item]
                    finally:
                        DEBUG and logger.info("run {{source}}", source=source)

                return BuiltFunction(get_schema_item, domain_schema[item], domain_schema)
            elif isinstance(item, FunctionFactory):
                fi, ti, si = _get(item, "build")(domain_type, domain_schema)

                def get_func_item(v, a):
                    try:
                        return f(v, a)[fi(v, a)]
                    finally:
                        DEBUG and logger.info("run {{source}}", source=source)

                return BuiltFunction(get_func_item, UnknownTyper(Exception("too complicated to know type")), s)
            else:

                def get_const_item(v, a):
                    try:
                        return f(v, a)[item]
                    finally:
                        DEBUG and logger.info("run {{source}}", source=source)

                return BuiltFunction(get_const_item, t[item], s)

        return FunctionFactory(builder, _get(self, "typer")[item], source)


    def __getattr__(self, item):
        source = f"{self}.{item}"

        def builder(domain_type, domain_schema) -> BuiltFunction:
            f, t, s = _get(self, "build")(domain_type, domain_schema)
            if item in s:

                def get_schema_item(v, a):
                    try:
                        return a[item]
                    finally:
                        DEBUG and logger.info("run {{source}}", source=source)

                return BuiltFunction(get_schema_item, domain_schema[item], domain_schema)
            elif isinstance(item, FunctionFactory):
                f, t, s = item.build(domain_type, domain_schema)

                def get_func_item(v, a):
                    try:
                        return getattr(f(v, a), f(v, a))
                    finally:
                        DEBUG and logger.info("run {{source}}", source=source)

                return BuiltFunction(get_func_item, UnknownTyper(Exception("too complicated to know type")), s)
            else:

                def get_const_item(v, a):
                    try:
                        return getattr(f(v, a), item)
                    finally:
                        DEBUG and logger.info("run {{source}}", source=source)

                return BuiltFunction(get_const_item, getattr(t, item), s)

        return FunctionFactory(builder, getattr(_get(self, "typer"), item), source)

    def __eq__(self, other):
        func_other = factory(other)

        def builder(domain_type, domain_schema) -> BuiltFunction:
            sf, st, ss = self.build(domain_type, domain_schema)
            of, ot, os = func_other.build(domain_type, domain_schema)

            def func(v, a):
                return sf(v, a) == of(v, a)

            return BuiltFunction(func, Typer(python_type=bool), domain_schema)

        return FunctionFactory(builder, Typer(python_type=bool), f"{other} == {self}")

    def __gt__(self, other):
        func_other = factory(other)

        def builder(domain_type, domain_schema) -> BuiltFunction:
            sf, st, ss = self.build(domain_type, domain_schema)
            of, ot, os = func_other.build(domain_type, domain_schema)

            def func(v, a):
                return sf(v, a) > of(v, a)

            return BuiltFunction(func, Typer(python_type=bool), domain_schema)

        return FunctionFactory(builder, Typer(python_type=bool), f"{other} > {self}")

    def __ge__(self, other):
        func_other = factory(other)

        def builder(domain_type, domain_schema) -> BuiltFunction:
            sf, st, ss = self.build(domain_type, domain_schema)
            of, ot, os = func_other.build(domain_type, domain_schema)

            def func(v, a):
                return sf(v, a) >= of(v, a)

            return BuiltFunction(func, Typer(python_type=bool), domain_schema)

        return FunctionFactory(builder, Typer(python_type=bool), f"{other} >= {self}")

    def __lt__(self, other):
        func_other = factory(other)

        def builder(domain_type, domain_schema) -> BuiltFunction:
            sf, st, ss = self.build(domain_type, domain_schema)
            of, ot, os = func_other.build(domain_type, domain_schema)

            def func(v, a):
                return sf(v, a) < of(v, a)

            return BuiltFunction(func, Typer(python_type=bool), domain_schema)

        return FunctionFactory(builder, Typer(python_type=bool), f"{other} < {self}")

    def __le__(self, other):
        func_other = factory(other)

        def builder(domain_type, domain_schema) -> BuiltFunction:
            sf, st, ss = self.build(domain_type, domain_schema)
            of, ot, os = func_other.build(domain_type, domain_schema)

            def func(v, a):
                return sf(v, a) <= of(v, a)

            return BuiltFunction(func, Typer(python_type=bool), domain_schema)

        return FunctionFactory(builder, Typer(python_type=bool), f"{other} <= {self}")

    def __truediv__(self, other):
        func_other = factory(other)

        def builder(domain_type, domain_schema) -> BuiltFunction:
            sf, st, ss = self.build(domain_type, domain_schema)
            of, ot, os = func_other.build(domain_type, domain_schema)

            def func(v, a):
                sv = sf(v, a)
                ov = of(v, a)
                if is_missing(sv) or is_missing(ov):
                    return None
                return sv / ov

            return BuiltFunction(func, Typer(python_type=float), domain_schema)

        return FunctionFactory(builder, Typer(python_type=float), f"{other} / {self}")

    def __rsub__(self, other):
        func_other = factory(other)

        def builder(domain_type, domain_schema) -> BuiltFunction:
            sf, st, ss = self.build(domain_type, domain_schema)
            of, ot, os = func_other.build(domain_type, domain_schema)

            def func(v, a):
                return of(v, a) - sf(v, a)

            return BuiltFunction(func, st, domain_schema)

        type_ = Typer(example=other) + _get(self, "typer")
        return FunctionFactory(builder, type_, f"{other} - {self}")

    def __add__(self, other):
        func_other = factory(other)

        def builder(domain_type, domain_schema) -> BuiltFunction:
            sf, st, ss = self.build(domain_type, domain_schema)
            of, ot, os = func_other.build(domain_type, domain_schema)

            def func(v, a):
                return sf(v, a) + of(v, a)

            return BuiltFunction(func, st+ot, domain_schema)

        type_ = _get(self, "typer") + _get(other, "typer")
        return FunctionFactory(builder, type_, f"{self} + {other}")

    def __radd__(self, other):
        func_other = factory(other)

        def builder(domain_type, domain_schema) -> BuiltFunction:
            sf, st, ss = self.build(domain_type, domain_schema)
            of, ot, os = func_other.build(domain_type, domain_schema)

            def func(v, a):
                return of(v, a) + sf(v, a)

            return BuiltFunction(func, st, domain_schema)

        type_ = Typer(example=other) + _get(self, "typer")
        return FunctionFactory(builder, type_, f"{other} + {self}")



    def __call__(self, *args, **kwargs):
        args = [factory(a) for a in args]
        kwargs = {k: factory(v) for k, v in kwargs.items()}

        desc_args = [str(a) for a in args]
        desc_args.extend(f"{k}={v}" for k, v in kwargs.items())
        params = ",".join(desc_args)
        source = f"{self}({params})"

        def builder(domain_type, domain_schema) -> BuiltFunction:
            sf, st, ss = self.build(domain_type, domain_schema)
            _args = BuiltFunction(*zip(*(a.build(st, ss) for a in args))) if args else NO_ARGS
            _kwargs = {k: v.build(st, ss).function for k, v in kwargs.items()}

            def func(v, a):
                callee = None
                result = None
                try:
                    callee = sf(v, a)
                    call_args = [f(v, a) for f in _args.function]
                    call_kwargs = {k: f(v, a) for k, f in _kwargs.items()}
                    result = callee(*call_args, **call_kwargs)
                    return result
                finally:
                    DEBUG and logger.info(
                        "call {{source}} on {{callee}} result {{result}}", source=source, callee=callee, result=result
                    )

            setattr(func, "source", source)

            return BuiltFunction(func, st(*_args.return_type), domain_schema)

        return FunctionFactory(builder, _get(self, "typer"), source)

    def __str__(self):
        return _get(self, "_desc")


def factory(item, return_type=None):
    if isinstance(item, FunctionFactory):
        return item

    # CONSTANT
    def build_constant(domain_type, domain_schema) -> BuiltFunction:
        return BuiltFunction(lambda v, a: item, Typer(example=item), domain_schema)

    return FunctionFactory(build_constant, Typer(example=item), f"{item}")


def normalize(item, return_type=None):
    """
    RETURN A FUNCTION THAT CAN ACT ON THE value/attachment PAIR FOUND IN STREAMS
    """
    if isinstance(item, FunctionFactory):
        return item

    if isinstance(item, (str, bytes, bool, int, float)):
        # CONSTANT
        def build_constant(domain_type, domain_schema) -> BuiltFunction:
            return BuiltFunction(lambda v, a: item, Typer(example=item), domain_schema)

        return FunctionFactory(build_constant, Typer(example=item), f"{item}")
    else:
        normalized_func, return_type = wrap_func(item, return_type=return_type)

        def builder(domain_type, domain_schema) -> BuiltFunction:
            if isinstance(return_type, LazyTyper):
                return BuiltFunction(normalized_func, domain_type, domain_schema)
            return BuiltFunction(normalized_func, return_type, domain_schema)

        return FunctionFactory(builder, return_type, f"returning {return_type}")


#
# def build(item) -> BuiltFunction:
#     if isinstance(item, FunctionFactory):
#         return item.build
#
#     def builder(domain_type, domain_schema) -> BuiltFunction:
#         return BuiltFunction(lambda v, a: item, Typer(example=item), domain_schema)
#
#     return builder


# build list of single arg builtins, that can be used as parse actions
singleArgBuiltins = [
    sum,
    len,
    sorted,
    reversed,
    list,
    tuple,
    set,
    any,
    all,
    min,
    max,
]

singleArgTypes = [
    int,
    float,
    str,
    bool,
    complex,
    dict,
]


def wrap_func(func, return_type=None):
    func_desc = str(func)
    if func_desc.startswith("<method"):
        func_name = strings.between(func_desc, "<method '", "' of '")
        class_name = strings.between(func_desc, "of '", "' objects")
        _class = first(t for t in singleArgTypes if t.__name__ == class_name)
        if _class:
            result = ANNOTATIONS.get((_class, func_name))
            if result:
                def wrapped(val, att):
                    return func(val)
                return wrapped, result()
    try:
        func_name = getattr(func, "__name__", getattr(func, "__class__").__name__)
    except Exception:
        func_name = func_desc
    if func_name.startswith("<"):
        func_name = "func"



    if func in singleArgBuiltins:
        spec = inspect.getfullargspec(func)
    elif func.__class__.__name__ == "staticmethod":
        func = func.__func__
        spec = inspect.getfullargspec(func)
    elif func.__class__.__name__ == "builtin_function_or_method":
        spec = inspect.getfullargspec(func)
    elif func in singleArgTypes:
        spec = inspect.FullArgSpec(["value"], None, None, None, [], None, {})
        return_type = func
    elif isinstance(func, type):
        spec = inspect.getfullargspec(func.__init__)
        new_func = func.__call__
        # USE ONLY FIRST PARAMETER
        num_args = len(spec.args) - 1  # ASSUME self IS FIRST ARG
        if num_args == 0:

            def wrap_init0(val, att):
                return new_func()

            return wrap_init0, Typer(python_type=func)
        else:

            def wrap_init1(val, att):
                return new_func(val)

            return wrap_init1, Typer(python_type=func)
    elif isinstance(func, FunctionType):
        spec = inspect.getfullargspec(func)
    elif hasattr(func, "__call__"):
        spec = inspect.getfullargspec(func)

    if spec.varargs:
        num_args = 3
    elif spec.args and spec.args[0] in ["cls", "self"]:
        num_args = len(spec.args) - 1
    else:
        num_args = len(spec.args)

    if not return_type:
        return_type = spec.annotations.get("return")
    if return_type:
        return_type = Typer(python_type=return_type)
    else:
        cause = Except(
            ERROR,
            "expecting {{function}} to have annotated return type",
            {"function": func_name},
            trace=get_stacktrace(start=3),
        )
        return_type = UnknownTyper(cause)

    if num_args == 0:

        def wrapper0(val, att):
            return func()

        wrapper = wrapper0
    elif num_args == 2 and spec.args[-1].startswith("att"):
        wrapper = func
    else:
        locals = {}
        exec(
            strings.outdent(
                f"""
            def {func_name}(val, att):
                return func(val)            
            """
            ),
            {"func": func},
            locals,
        )
        wrapper = locals[func_name]
        setattr(wrapper, "original", func)

    # copy func name to wrapper for sensible debug output
    wrapper.__name__ = func_name
    return wrapper, return_type


class TopFunctionFactory(FunctionFactory):
    """
    it(x)  RETURNS A FunctionFactory FOR x
    """

    def __call__(self, value):
        if isinstance(value, FunctionFactory):
            logger.error("don't do this")

        if isinstance(value, type):
            # ASSUME THIS IS A CONSTRUCTOR
            typer = CallableTyper(return_type=value)

            def type_builder(domain_type, domain_schema) -> BuiltFunction:
                return BuiltFunction(lambda v, a: value, typer, domain_schema)

            return FunctionFactory(type_builder, typer, f"{value}")

        typer = Typer(python_type=type(value))

        def value_builder(domain_type, domain_schema) -> BuiltFunction:
            return BuiltFunction(lambda v, a: value, typer, domain_schema)

        return FunctionFactory(value_builder, typer, f"{value}")

    def __str__(self):
        return "it"


def noop(v, a):
    return v


def top_builder(domain_type, domain_schema) -> BuiltFunction:
    return BuiltFunction(noop, domain_type, domain_schema)


it = TopFunctionFactory(top_builder, LazyTyper(), "it")
